#define VERSION "1.3.2"
/*
* Author  : Yury Pimenov (http://up.ftpsearch.lv/)
* Mailto  : up@fs.lv
* Desc    : The traffic accounting daemon with
*           filter of latvian and foreign nets
*           and logging to MySQL database. See
*           the README.
*
* History :
* Version 1.0
*     18.07.01 Initial creation
* Version 1.1
*     31.07.01 flush() fork()s for fastest update
*     31.07.01 [-l loglevel] switch added
* Version 1.3.1
*     12.02.01 feature: -d debug flag added
* Version 1.3.2
*     18.07.02 cleaned code for publication
*
*/

#include "trcounterd.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <mysql.h>
#include <pcap.h>

struct user_info {
    struct in_addr ip;
    int latvian_in;
    int latvian_out;
    int foreign_in;
    int foreign_out;
    int ps_recv_in;
    int ps_recv_out;
};

struct net {
    struct in_addr network;
    struct in_addr netmask;
};

struct net_list {
    struct net *first;
    int num;
};

// global variables

FILE *logfile;
struct user_info *acc = NULL;
struct net inner_net;
struct net_list lv_nets;
struct net_list local_nets;
time_t last_flush;
u_int ps_recv = 0;
int debug_mode = 0;

#define is_from_net(A,B) (( (B).s_addr & (A).netmask.s_addr ) == (A).network.s_addr )
#define atoul(A) (strtoul(A, (char **) NULL, 10))

#define LOG_ERROR 1
#define LOG_WARNING 2
#define LOG_INFO 3
#define LOG_DEBUG 4

int num_of_users;
pcap_t *ip_socket;
int dlt_len = 0;
int loglevel = LOG_DEBUG;

#define HAVE_STDARG_H 1

void 
#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
logz( int level, char *fmt, ... )
#else
logz( int level, va_alist )
va_dcl
#endif
{
    va_list ap;
#if !(defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__)
    const char *fmt;
#endif
    time_t now_sec;
    struct tm *now;

    if ( level > loglevel )
        return;

#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
        va_start ( ap, fmt );
#else
        va_start( ap );
        fmt = va_arg( ap, char * );
#endif

    now_sec = time(NULL);
    now = localtime(&now_sec);

    fprintf( logfile, "%.2d/%.2d/%.2d %.2d:%.2d:%.2d [%d] ",
        now->tm_mday, now->tm_mon+1, now->tm_year-100,
        now->tm_hour, now->tm_min, now->tm_sec, getpid() );
           
    switch( level ) {
    case LOG_ERROR :
        fprintf( logfile, "Error: ");
        break;
    case LOG_WARNING :
        fprintf( logfile, "Warning: ");
        break;
    case LOG_INFO :
        fprintf( logfile, "Info: ");
        break;
    case LOG_DEBUG :
        fprintf( logfile, "Debug: ");
        break;
    }
    
    vfprintf( logfile, fmt, ap );
    fprintf( logfile, "\n" );
    fflush( logfile );

    va_end( ap );
}


int is_from_net_list( struct net_list net_list, struct in_addr ip)
{
    int num;
    struct net *first;

    num = net_list.num;
    first = net_list.first;

    while( num-- )
        if ( is_from_net( first[num], ip ) )
            return 1;

    return 0;
}

// returns a number of hosts in network
// or -1 if error
int net_aton( char *c_net, struct net *sn )
{
    char *c_mask;

    c_mask = strchr( c_net, '/' );

    if ( c_mask ) {
        int mask_bits;

        *c_mask = '\0';
        c_mask++;
        if ( strchr( c_mask, '.' ) )
            inet_aton( c_mask, &sn->netmask );
        else {
            if ( (mask_bits = atol( c_mask )) > 32 )
                return -1;
            sn->netmask.s_addr = htonl(~((1 << ( 32 - mask_bits )) - 1));
        }
    }
    else
        sn->netmask.s_addr = 0xFFFFFFFF;

    inet_aton( c_net, &sn->network );

    return ~ntohl( sn->netmask.s_addr ) + 1 ;
}

// 0 = error, !0 = ok
int netlist_load(char *fn, struct net_list *nl)
{
    FILE *f;
    int num = 0;
    char buf[128];
    struct net *first = NULL;
    char *hash;
    struct net sn;

    f = fopen( fn, "r" );
    if ( !f )
        return 0;
    fgets( buf, sizeof( buf ), f );
    while( !feof(f) ) {
        hash = strchr( buf, '#' );
        if ( hash )
            *hash = '\0';
        if ( ( !*buf ) || ( net_aton( buf, &sn ) == -1 ) )
            goto nextline;
        first = realloc( first, sizeof( struct net ) * (num + 1));
        first[num] = sn;
        num++;
nextline:
        fgets( buf, sizeof( buf ), f );
    }
    fclose( f );

    nl->first = first;
    nl->num = num;
}

MYSQL* connect_db()
{
    static MYSQL handle;

    if ( ! mysql_connect( &handle, MYSQL_HOST, MYSQL_USER, MYSQL_PASS ) ) {
        logz( LOG_ERROR, "mysql_connect(): unable to connect MySQL\n" );
        return NULL;
    }
    if ( mysql_select_db( &handle, MYSQL_DB ) == -1 ) {
        logz( LOG_ERROR, "mysql_select_db(): unable to use database " MYSQL_DB "\n" );
        return NULL;
    }
    return &handle;
}

void kill_zombies( int signum )
{
    int status;

    while( waitpid( -1, &status, WNOHANG ) > 0 ) 
        if ( WEXITSTATUS( status ) )
            logz( LOG_WARNING, "flushing failed, collected data droped" );
}

void flush( int signum )
{
    MYSQL *handle;
    MYSQL_ROW row;
    struct user_info *cur;
    int i;
    char query[512];
    struct pcap_stat stats;
    int rval;

    logz( LOG_INFO, "flush(): packets received %d", ps_recv );
    ps_recv = 0;
    pcap_stats( ip_socket, &stats );
//    logz( LOG_INFO, "pcap_stats(): totaly packets received %d, dropped %d", stats.ps_recv, stats.ps_drop );

    if ( ! fork() ) {
        if ( ! ( handle = connect_db() ) )
            exit( 1 );
        rval = 0;
        cur = acc - 1;
        i = num_of_users;
        while ( cur++, i-- ) {
            if ( cur->latvian_in || cur->latvian_out || cur->foreign_in || cur->foreign_out ) {
                logz( LOG_DEBUG, "Ip: %s In (%d, %d) Out (%d, %d)", inet_ntoa( cur->ip ),
                    cur->ps_recv_in, cur->foreign_in + cur->latvian_in,
                    cur->ps_recv_out, cur->foreign_out + cur->latvian_out );
                sprintf( query, "UPDATE " USERS_TABLE " SET current_latvian=current_latvian+%d,"
                    "current_foreign=current_foreign+%d WHERE ip=%lu",
                    cur->latvian_in + cur->latvian_out, cur->foreign_in + cur->foreign_out, ntohl( cur->ip.s_addr ) );
                if ( mysql_query( handle, query ) == -1 ) {
                    logz( LOG_ERROR, "mysql_query(): unable to execute query: %s\n" );
                    rval++;
                    break;
                }
                sprintf( query, "INSERT INTO " LOG_TABLE " SET ip=%lu,time_from=FROM_UNIXTIME(%d),time_until=NOW(),"
                    "in_latvian=%d,out_latvian=%d,in_foreign=%d,out_foreign=%d",
                    ntohl( cur->ip.s_addr ), last_flush,
                    cur->latvian_in, cur->latvian_out, cur->foreign_in, cur->foreign_out );
                if ( mysql_query( handle, query ) == -1 ) {
                    logz( LOG_ERROR, "mysql_query(): unable to execute query: %s\n" );
                    rval++;
                    break;
                }
            }
        }
        free( acc );
        exit( rval );
    }

    load_config();

    last_flush = time( NULL );

//    pcap_close(ip_socket);

    return;
}

void flush_n_exit( int signum )
{
    flush(0);
    fclose( logfile );
    exit(0);
}

void reload( int signum )
{
    /* reopen log */
    fclose( logfile );
    if ( ! (logfile = fopen( LOG_FILE, "a" ) ) ) {
        logfile = stderr;
        logz( LOG_WARNING, "fopen(): unable to open log file " LOG_FILE "\n" );
    }

    /* reload config */
    load_config();
}

void filter_packet(u_char *u, struct pcap_pkthdr *p, u_char *packet)
{
    #define IP_SIZE	20
    struct iphdr *ip;
    struct user_info *cur;
    struct net sn;
    int i;

    if ( p->len < (dlt_len + IP_SIZE ) )
        return;
    ip = (struct iphdr *) ( packet + dlt_len ) ;

    if ( debug_mode ) {
        printf( "%s ", inet_ntoa( *((struct in_addr*) &ip->saddr) ) );
        printf( "%s %d ", inet_ntoa( *((struct in_addr*) &ip->daddr) ), 
            ntohs( ip->tot_len ) );
    }

    // dst and src not from inner_net
    if ( (!is_from_net( inner_net, *((struct in_addr*) (&ip->saddr)) )) &&
        (!is_from_net( inner_net, *((struct in_addr*) (&ip->daddr)) )) ) {
        if ( debug_mode ) printf( "EXTERNAL > EXTERNAL\n" );
        return;
    }
    // Inner -> Inner
    if ( is_from_net( inner_net, *((struct in_addr*) (&ip->saddr)) ) &&
        is_from_net( inner_net, *((struct in_addr*) (&ip->daddr)) ) ) {
        if ( debug_mode ) printf( "INTERNAL > INTERNAL\n" );
        return;
    }
    // Local -> Local
    if ( is_from_net_list( local_nets, *((struct in_addr*) (&ip->saddr)) ) &&
        is_from_net_list( local_nets, *((struct in_addr*) (&ip->daddr)) ) ) {
        if ( debug_mode ) printf( "LOCAL > LOCAL\n" );
        return;
    }

    ps_recv++;

    i = num_of_users; cur = acc;
    // Inner -> Latvian or Foreign
    if ( is_from_net( inner_net, *((struct in_addr*) (&ip->saddr)) ) ) {
        while( i && ( cur->ip.s_addr != ip->saddr ) ) {
            cur++;
            i--;
        }
        if ( !i )
            return;

        cur->ps_recv_out++;

        if ( debug_mode ) printf( "INTERNAL > " );

        if ( is_from_net_list( lv_nets, *((struct in_addr*) (&ip->daddr)) ) ) {
            if ( debug_mode ) printf( "LATVIAN\n" );
            cur->latvian_out += ntohs( ip->tot_len );
        }
        else {
            if ( debug_mode ) printf( "FOREIGN\n" );
            cur->foreign_out += ntohs( ip->tot_len );
        }
        return;
    }
    // Latvian or Foreign -> Inner
    if ( is_from_net( inner_net, *((struct in_addr*) (&ip->daddr)) ) ) {
        while( i && ( cur->ip.s_addr != ip->daddr ) ) {
            cur++;
            i--;
        }
        if ( !i )
            return;

        cur->ps_recv_in++;

        if ( is_from_net_list( lv_nets, *((struct in_addr*) (&ip->saddr)) ) ) {
            if ( debug_mode ) printf( "LATVIAN > " );
            cur->latvian_in += ntohs( ip->tot_len );
        }
        else {
            if ( debug_mode ) printf( "FOREIGN > " );
            cur->foreign_in += ntohs( ip->tot_len );
        }
        if ( debug_mode ) printf( "INTERNAL\n" );
        return;
    }
}

int load_config()
{
    MYSQL *handle;
    MYSQL_RES *result;
    MYSQL_ROW row;
    char query[512];
    struct user_info *cur;

    netlist_load( LATVIAN_NETS_LIST, &lv_nets );
    netlist_load( LOCAL_NETS_LIST, &local_nets );

    if (! (handle = connect_db()) )
        return 0;

    sprintf( query, "SELECT ip FROM " USERS_TABLE );
    if ( mysql_query( handle, query ) == -1 ) {
        logz( LOG_ERROR, "mysql_query(): unable to execute query %s", query );
        return 0;
    }
    result = mysql_store_result( handle );
    if ( ! ( num_of_users = mysql_num_rows( result ) ) )
        logz( LOG_WARNING, "mysql_num_rows(): " USERS_TABLE " table is empty" );

    if ( acc )
        free( acc );

    cur = acc = calloc( num_of_users, sizeof( struct user_info ) ); // calloc initialize the memory with 0

    while ( row = mysql_fetch_row( result ) ) {
        cur->ip.s_addr = htonl( atoul( row[0] ) );
        cur++;
    }

    mysql_close( handle );
    return 1;
}

void finish_debug_mode( int signum )
{
    printf( "Exiting...\n" );
    exit( 0 );
}

void usage( char *argv_0 )
{
    printf( "%s [-h] [-d] [-p] [-i interface] [-f filter] [-l loglevel] \n", argv_0 );
    printf( "trcounterd v" VERSION "\n" );
    printf( "written by Yury Pimenov <up@fs.lv> (c) 2002\n" );
    printf( "    -h this help\n" );
    printf( "    -d turns on debug mode\n" );
    printf( "    -p set promiscious mode\n" );
    printf( "    -i select interface\n" );
    printf( "    -l select loglevel (0-no output ... 4-maximum verbose). Default %d\n", loglevel );
    printf( "    -f set filter for pcap library. Default \"" FILTERPROG "\"\n" );
    printf( "\n" );
    exit( 0 );                     
}

int main( int argc, char **argv )
{
    char *interface=NULL;
    char *myfilter=FILTERPROG;
    char errbuf[PCAP_ERRBUF_SIZE];
    struct bpf_program prog;
    int c;
    int promisc = 1;
    extern int opterr;   
    extern char *optarg;
    int local_net_size;
    int pid;
    FILE *pidfile;
    struct sigaction action, old_action;

    opterr = 0;   
    while( ( c = getopt( argc, argv, "hdpi:l:f:") ) != -1 )
        switch( c ) {
        case 'h':
            usage( argv[0] );
            break; /* never reach this line :) */
        case 'd':
            debug_mode++;
            break;
        case 'i':
            interface = optarg;
            break;
        case 'p':
            promisc = 0;
            break;
        case 'f':
            myfilter = optarg;
            break;
        case 'l':
            loglevel = atoi( optarg );
            break;
    }

    logfile = stderr;

    if ( interface == NULL ) {
        interface = pcap_lookupdev( errbuf );
        if ( interface == NULL ) {
            logz( LOG_ERROR, "pcap_lookupdev(): %s", errbuf );
            exit( 1 );
        }
    }
    if ( pcap_lookupnet( interface, (bpf_u_int32*) &(inner_net.network.s_addr),
        (bpf_u_int32*) &(inner_net.netmask.s_addr), errbuf) < 0 ) {
        logz( LOG_ERROR, "pcap_lookupnet(): %s", errbuf );
        exit( 1 );
    }

    ip_socket = pcap_open_live( interface, 1024, promisc, 1024, errbuf );
    if ( ip_socket == NULL ) {
        logz( LOG_ERROR, "pcap_open_live(): %s", errbuf );
        exit( 1 );
    }
    switch( pcap_datalink( ip_socket ) ) {
        case DLT_EN10MB:      
            dlt_len = 14;
            break;
        default:
            logz( LOG_ERROR, "unknown datalink" );
            exit( 1 );
    }
    if ( pcap_compile( ip_socket, &prog, myfilter, 1, inner_net.netmask.s_addr ) < 0 ) {
        logz( LOG_ERROR, "pcap_compile(): %s", errbuf );
        exit( 1 );
    }
    if ( pcap_setfilter( ip_socket, &prog ) < 0 ) {
        logz( LOG_ERROR, "pcap_setfilter(): %s", errbuf );
        exit( 1 );
    }   

    if ( ! load_config() ) {
        logz( LOG_ERROR, "load_config(): " );
        exit( 1 );
    }

    if ( ! debug_mode )
        if ( pid = fork() ) {
            if ( pid == -1 ) {
                logz( LOG_ERROR, "fork(): unable to become a daemon\n" );
                exit( 1 );
            }
            if ( ! (pidfile = fopen( PID_FILE, "w" ) ) ) {
                logz( LOG_ERROR, "fopen(): unable to open pid file " PID_FILE "\n" );
                 exit( 1 );
            }
            fprintf( pidfile, "%d", pid );
            fclose( pidfile );
            exit( 0 );
        }

    last_flush = time( NULL );

    sigemptyset( &action.sa_mask );
    action.sa_handler = flush;
    action.sa_flags = 0;
    if ( sigaction( SIGUSR1, &action, &old_action ) == -1 ) {
        logz( LOG_ERROR, "sigaction()" );
        exit( 1 );
    }
    sigemptyset( &action.sa_mask );
    action.sa_handler = flush_n_exit;
    action.sa_flags = 0;
    if ( sigaction( SIGTERM, &action, &old_action ) == -1 ) {
        logz( LOG_ERROR, "sigaction()" );
        exit( 1 );
    }
    sigemptyset( &action.sa_mask );
    action.sa_handler = kill_zombies;
    action.sa_flags = 0;
    if ( sigaction( SIGCHLD, &action, &old_action ) == -1 ) {
        logz( LOG_ERROR, "sigaction()" );
        exit( 1 );
    }
    sigemptyset( &action.sa_mask );
    action.sa_handler = reload;
    action.sa_flags = 0;
    if ( sigaction( SIGHUP, &action, &old_action ) == -1 ) {
        logz( LOG_ERROR, "sigaction()" );
        exit( 1 );
    }

    if ( debug_mode ) {
        sigemptyset( &action.sa_mask );
        action.sa_handler = finish_debug_mode;
        action.sa_flags = 0;
        if ( sigaction( SIGINT, &action, &old_action ) == -1 ) {
            logz( LOG_ERROR, "sigaction()" );
            exit( 1 );
        }
    }
    else {
        sigemptyset( &action.sa_mask );
        action.sa_handler = flush;
        action.sa_flags = 0;
        if ( sigaction( SIGUSR1, &action, &old_action ) == -1 ) {
            logz( LOG_ERROR, "sigaction()" );
            exit( 1 );
        }
        sigemptyset( &action.sa_mask );
        action.sa_handler = flush_n_exit;
        action.sa_flags = 0;
        if ( sigaction( SIGTERM, &action, &old_action ) == -1 ) {
            logz( LOG_ERROR, "sigaction()" );
            exit( 1 );
        }
        sigemptyset( &action.sa_mask );
        action.sa_handler = kill_zombies;
        action.sa_flags = 0;
        if ( sigaction( SIGCHLD, &action, &old_action ) == -1 ) {
            logz( LOG_ERROR, "sigaction()" );
            exit( 1 );
        }
        sigemptyset( &action.sa_mask );
        action.sa_handler = reload;
        action.sa_flags = 0;
        if ( sigaction( SIGHUP, &action, &old_action ) == -1 ) {
            logz( LOG_ERROR, "sigaction()" );
            exit( 1 );
        }

        if ( ! ( logfile = fopen( LOG_FILE, "a" ) ) ) {
            logfile = stderr;
            logz( LOG_WARNING, "fopen(): unable to open log file " LOG_FILE "\n" );
        }
    }

    while( 1 )
        pcap_loop( ip_socket, -1, (pcap_handler) filter_packet, NULL );

}
